//===============================================================================
//            55Lava                
//            by Kazuto Saiki
//         2018.09.25 version  
//        2020.08.03 version   
//===============================================================================

//  ユーザー定義定数
final int X_SIZE = 1026;     // 地図のサイズ(pixels)
final int Y_SIZE = 1026;     // 地図のサイズ(pixels)
final int X_MIN = 1;        // 計算範囲を地図範囲より絞る時のX最小値 >=1
final int X_MAX = 1025;      //計算範囲を地図範囲より絞る時のX最大値 <X_SIZE-1
final int Y_MIN = 1;        // 計算範囲を地図範囲より絞る時のY最小値 >=1
final int Y_MAX = 1025;     //計算範囲を地図範囲より絞る時のY最大値 <Y_SIZE-1
final float Density = 2700;     // 溶岩の密度 (kg/m3)
final float Cp = 1200;        // 溶岩の比熱 (J/kg/K)
final float G = 9.8;         // 重力加速度　　(m/s2)
final float TimeInterval = 30.0;   // 結果画像をファイルに描き出す時間インターバル (s)
final float CellSize = 30.0;        //セルのサイズ　（m）
final float RoomTemp = 293.15;      //室温（K)
final float MagmaTemp = 1373.15;     //溶岩の噴出時の初期温度(K)


float [][] terrain = new float[X_SIZE][Y_SIZE];     //　地形標高  (m)
float [][] lava = new float[X_SIZE][Y_SIZE];        //  溶岩流厚さ (m)
float [][] delta_lava = new float[X_SIZE][Y_SIZE];  //  そのセルへの溶岩流入量（深さ換算）　(m)
float [][] minus_delta_lava = new float[X_SIZE][Y_SIZE]; //  周辺セルへの溶岩流出量（深さ換算）　(m)
float [][] q = new float[X_SIZE][Y_SIZE];       //  そのセルのマグマ熱量 (J)
float [][] delta_q = new float[X_SIZE][Y_SIZE];       //  セルへのマグマ熱量流入量 (J)
float [][] minus_delta_q = new float[X_SIZE][Y_SIZE];  // 周辺セルへのマグマ熱量流出量　（J)
float time;   // シミュレーション経過時間(s)
float timer;   //フレーム書き出しのためタイマー
int counter;   //フレーム書き出しファイルのナンバリング用

int cyc;
PImage img;

/*-------------------------------------------------------------------------------------------------*/
void settings(){
  size(X_SIZE,Y_SIZE+100);
}

void setup() {
  colorMode(RGB);
  frameRate(10000);

//  地図の読み込み
  img=loadImage("map.png");
  image(img,0,0);
  time = 0.0;
  timer= -1.0;
  counter=0;

  for (int y=0; y<Y_SIZE; y++)
      for (int x=0; x<X_SIZE; x++){
        color c = get(x,y);
        terrain[x][y]=green(c)*16.0+blue(c);
        if(blue(c)==255) terrain[x][y]=0.0;
        lava[x][y]=0.0;
        delta_lava[x][y]=0.0;
        minus_delta_lava[x][y]=0.0;        
        q[x][y]=0.0;
        delta_q[x][y]=0.0;       
        minus_delta_q[x][y]=0.0;    
      }
}

 /*-------------------------------------------------------------------------------------------------*/
 float flux(float h,float dz, float dh, float sy,float vis)
 {
   float result;
   float hcr; 
   float a;
   
   if((dz+dh)!=0) //<>//
     hcr = sy*sqrt(dz*dz+30*30)/(Density*G*(dz+dh)); 
   else return(0);
   
   a = h/hcr;
   if(a<=1) result = 0; 
   else result = sy*hcr*hcr*30*(a*a*a-1.5*a*a+0.5)/3/vis/(CellSize*CellSize);
   
   return(result);
 }
   
 /*-------------------------------------------------------------------------------------------------*/
void draw(){
  float north,east,west,south;
  float sum;
  float dz,dh,sy,vis;
  float org_h, des_h;
  float org_l, des_l;
  float time_mag;
  int mx,my;
  int elevation;
  float temp;
  float temp_t;
  float dq;
 
  //  マウスによる各種情報表示
  fill(0, 0, 255);
  stroke(0,0,255);
  rect(0,Y_SIZE,X_SIZE,Y_SIZE+200);
                       //ゆっくりマウスを外に出すと表示が消えるように、反応領域を1pixel削る
  if((mouseX>1)&&(mouseX<(X_SIZE-1))&&
        (mouseY>1)&&(mouseY<(Y_SIZE-1))){
    color c = get(mouseX, mouseY);
    elevation = (int)(green(c)*16+blue(c));
    if(blue(c)==255) elevation = 0;
 
    fill(255,255,255);
    textSize(25);
    text("  X="+mouseX+
    "  Y="+mouseY+"   ELEVATION="+elevation+"m", 0, Y_SIZE+30);
    if(lava[mouseX][mouseY]>0)
      text("   Lava Thickness="+nf(lava[mouseX][mouseY],3,3)+"m    Temp."
      +nf((q[mouseX][mouseY]/(lava[mouseX][mouseY]*(CellSize*CellSize)*Density*Cp)-273.15),4,0)
      +"  Q ="+q[mouseX][mouseY]
     , 0, Y_SIZE+70);
  }

  cyc++;
  time_mag=1.0;

  // flux calculation  流れ出すのが正
  for (int y=Y_MIN; y<Y_MAX; y++){
      for (int x=X_MIN; x<X_MAX; x++){

        if(lava[x][y]==0.0){
          north = 0.0; 
          east = 0.0;
          west = 0.0;
          south = 0.0;
      }
      else{ 
          temp = q[x][y]/(lava[x][y]*(CellSize*CellSize)*Density*Cp);
          vis =exp(2.307*(10000/temp-1.5)-8.7);
          sy = pow(10,(13.00997-0.0089*temp));
          org_h=terrain[x][y];
          org_l=lava[x][y];
          
        // North   
          des_h=terrain[x][y-1];
          des_l=lava[x][y-1];         
          dh = org_l-des_l;
          dz = org_h-des_h;           
          north = flux(org_l,dz,dh,sy,vis);
       
        //East
          des_h=terrain[x+1][y];
          des_l=lava[x+1][y];         
          dh = org_l-des_l;
          dz = org_h-des_h;         
          east = flux(org_l,dz,dh,sy,vis); 
        
        //West
          des_h=terrain[x-1][y];
          des_l=lava[x-1][y];         
          dh = org_l-des_l;
          dz = org_h-des_h;         
          west = flux(org_l,dz,dh,sy,vis);
        
        //South
          des_h=terrain[x][y+1];
          des_l=lava[x][y+1];         
          dh = org_l-des_l;
          dz = org_h-des_h;        
          south = flux(org_l,dz,dh,sy,vis);

          delta_lava[x][y-1] += north;
          delta_lava[x+1][y] += east;
          delta_lava[x-1][y] += west;
          delta_lava[x][y+1] += south;
          
          delta_q[x][y-1] += north/lava[x][y]*q[x][y];
          delta_q[x+1][y] += east/lava[x][y]*q[x][y];
          delta_q[x-1][y] += west/lava[x][y]*q[x][y];
          delta_q[x][y+1] += south/lava[x][y]*q[x][y];
          
          sum = north+east+west+south;
          minus_delta_lava[x][y] = sum;
          minus_delta_q[x][y] = sum/lava[x][y]*q[x][y];
          if((sum > lava[x][y]) && (time_mag < sum/lava[x][y]))
             time_mag = sum / lava[x][y];

      }
    }
  }
    
    for (int y=Y_MIN; y<Y_MAX; y++){
      for (int x=X_MIN; x<X_MAX; x++){           
        lava[x][y]= lava[x][y]+(delta_lava[x][y]-minus_delta_lava[x][y])/time_mag;
        if(lava[x][y]<0) lava[x][y]=0.0;
        delta_lava[x][y]=0.0;
        minus_delta_lava[x][y]=0.0;
        
        q[x][y]= q[x][y]+(delta_q[x][y]-minus_delta_q[x][y])/time_mag;
        if(q[x][y]<0) q[x][y]=0.0;
        delta_q[x][y]=0.0;
        minus_delta_q[x][y]=0.0;       
  
        //　マグマ表示
        if(lava[x][y]>0.0){
              color c = get(x, y);
              stroke(255,green(c),blue(c));
              point(x,y);
        }
      }
    }
    time = time+1/time_mag;
    println(cyc+" TimeMag  "+time_mag+"  time"+time);
    
    //冷却プロセス  
    for (int y=Y_MIN; y<Y_MAX; y++){
      for (int x=X_MIN; x<X_MAX; x++){
        if(lava[x][y]>0){
          temp_t = q[x][y]/(lava[x][y]*(CellSize*CellSize)*Density*Cp);
          dq = 5.67*(CellSize*CellSize)*(pow(temp_t,4)-pow(RoomTemp,4))/pow(10,8)/time_mag;
          q[x][y]=q[x][y]-dq;
          if(q[x][y]/(lava[x][y]*(CellSize*CellSize)*Density*Cp)<RoomTemp)
              q[x][y]=RoomTemp*(lava[x][y]*(CellSize*CellSize)*Density*Cp);
        }
      }
    }
  
    
    timer = timer -1/time_mag;
    if(timer < 0.0){
      save("frame/snap"+nf(counter,5)+".png");     
      timer = TimeInterval;             //TimeInterval秒ごとに記録
      counter++;
    }
    
    //   溶岩噴出　　噴出セル座標とセルあたりのフラックス　（高さ換算）（ｍ）で指定
    lava[812][665]+=20.0/time_mag;                                      //  += が　＝＋になっていたのを修正しました。2020/08/03
    q[812][665]=lava[812][665]*(CellSize*CellSize)*Density*Cp*MagmaTemp;
}
